#!/usr/local/bin/lua


print"-- Redirected output"
do
	local function with_output(f, func, ...)
		finally(io.output(), io.output)
		io.output(f)
		return func(...)
	end

	local f = io.tmpfile()
	local s = "test redirect"
	with_output(f, io.write, s)
	io.write"Display\n"
	f:seek("set")
	assert(f:read("*a") == s)
	f:close()
	print"OK"
end

print"-- Finalize on error"
do
	local function add5(x)
		finally(x, print, false)
		return x + 5
	end

	add5(5)
	assert(not pcall(add5), "Error expected")
	print"OK"
end

print"-- Manually finalized block"
do
	local nproc = 2

	local obj = {v = 0}
	function obj:close(e)
		assert(not e or e == self, "Error object expected")
		self.v = self.v + 1
	end

	local function block()
		for i = 1, nproc do
			local fin = finally(obj)
			if i == nproc then error(obj) end
			finalize(fin)
		end
	end

	pcall(block)
	assert(obj.v == nproc, "Object not processed")
	print"OK"
end

print"-- Tail call"
do
	local nproc = 2

	local obj = {v = ""}
	function obj:close() self.v = self.v .. "c" end

	local i = nproc

	local function tail()
		finally(obj)
		obj.v = obj.v .. "p"
		i = i - 1
		if i > 0 then return tail() end
	end

	tail()
	assert(obj.v == string.rep("pc", nproc))
	print"OK"
end

print"-- Transaction"
do
	local str = ""

	local function undo(step)
		str = str .. step
	end

	local function transact()
		local f = "Foo"
		finally(f, undo, false)

		local b = "Bar"
		finally(b, undo, false)

		local d = error("Def")

		return print(f, b, d)
	end

	pcall(transact)
	assert(str == "BarFoo", "Failure expected")
	print"OK"
end

